from ariadne_extended.resolvers import ListModelResolver, Resolver, DetailModelMixin
from waffle import get_waffle_flag_model
from waffle.models import Sample, Switch
from waffle.utils import get_setting

from .types import query, waffle_item_interface, waffle_type


class WaffleResolver(DetailModelMixin, ListModelResolver):
    lookup_arg = "name"
    lookup_field = "name"

    def get_object(self):
        lookup = self.get_lookup_filter_kwargs()
        # Waffle model get returns an empty class if not found with the used lookup name
        return self.model.get(**lookup)

    def get_queryset(self):
        # Not a true QS, will ask cache for saved model objects
        return self.model.get_all()


class FlagResolver(WaffleResolver):
    model = get_waffle_flag_model()


class SwitchResolver(WaffleResolver):
    model = Switch


class SampleResolver(WaffleResolver):
    model = Sample


class AllWaffleTypesResolver(Resolver):
    def list(self, *args, **kwargs):
        return Switch.get_all() + Sample.get_all() + get_waffle_flag_model().get_all()


class ActiveResolver(Resolver):
    def check(self, parent, *args, **kwargs):
        try:
            return parent.is_active(self.request)
        except TypeError:
            return parent.is_active()


waffle_type.set_field("flags", FlagResolver.as_resolver(method="list"))
waffle_type.set_field("switches", SwitchResolver.as_resolver(method="list"))
waffle_type.set_field("samples", SampleResolver.as_resolver(method="list"))
waffle_type.set_field("all", AllWaffleTypesResolver.as_resolver(method="list"))
waffle_type.set_field("flag", FlagResolver.as_resolver(method="retrieve"))
waffle_type.set_field("switch", SwitchResolver.as_resolver(method="retrieve"))
waffle_type.set_field("sample", SampleResolver.as_resolver(method="retrieve"))
waffle_item_interface.set_field("active", ActiveResolver.as_resolver(method="check"))
query.set_field(
    "waffle",
    lambda *x: {
        "flag_default": get_setting("FLAG_DEFAULT"),
        "switch_default": get_setting("SWITCH_DEFAULT"),
        "sample_default": get_setting("SAMPLE_DEFAULT"),
        "flagDefault": get_setting("FLAG_DEFAULT"),
        "switchDefault": get_setting("SWITCH_DEFAULT"),
        "sampleDefault": get_setting("SAMPLE_DEFAULT"),
    },
)



import logging
from importlib import import_module

from ariadne import EnumType, ObjectType, ScalarType, UnionType, load_schema_from_path
from django.apps import AppConfig
from django.apps.registry import apps

logger = logging.getLogger(__name__)


class GraphLoaderConfig(AppConfig):
    """
    Graph loader finds all the schemas, resolvers and type definitions in installed apps
    """

    name = "ariadne_extended.graph_loader"

    def ready(self):
        # Find all schema.graphql and resolver modules and load them
        self.type_defs = list()
        self.types = list()
        self.resolvers = list()

        logger.debug("Trying to load resolvers and schemas:")
        for config in apps.get_app_configs():
            self.load_schema(config)
            self.load_resolvers(config)
            self.load_custom_types(config)

        logger.debug("Types: %s", ", ".join([_type.name for _type in self.types]))

    def load_schema(self, config):
        try:
            schema = load_schema_from_path(config.path)
            if schema:
                self.type_defs.append(schema)
                logger.debug("found schema in %s!", config.name)
        except FileNotFoundError:
            pass

    def load_resolvers(self, config):
        try:
            module = import_module(".resolvers", package=config.name)
            resolvers = getattr(module, "__resolvers__", None)
            if resolvers:
                self.resolvers += [getattr(module, resolver) for resolver in resolvers]
        except ModuleNotFoundError as e:
            if e.name != f"{config.name}.resolvers":
                raise e
        except Exception as e:
            logger.error("Error loading types from %s", config.name)
            raise e
        else:
            logger.debug("found resolvers in %s!", config.name)

    def load_custom_types(self, config):
        try:
            types_module = import_module(".types", package=config.name)
            for type_name in [
                t
                for _, t in types_module.__dict__.items()
                if isinstance(t, ObjectType)
                or isinstance(t, ScalarType)
                or isinstance(t, EnumType)
                or isinstance(t, UnionType)
            ]:
                self.types.append(type_name)
        except ModuleNotFoundError as e:
            if e.name != f"{config.name}.types":
                raise e
        except Exception as e:
            logger.error("Error loading types from %s", config.name)
            raise e
        else:
            logger.debug("found types in %s!", config.name)

    @property
    def all_app_types(self):
        return self.types + self.resolvers



"""
Mixins that are in DRF serializers and resolver data
"""
import enum

from django.db.models.deletion import IntegrityError, ProtectedError
from . import exceptions


class ListModelMixin:
    """
    List a queryset.
    """

    pagination_class = None

    def list(self, request, *args, **kwargs):
        queryset = self.get_queryset()

        page = self.paginate_queryset(queryset)
        if page is not None:
            return page

        return queryset

    @property
    def paginator(self):
        """
        The paginator instance associated with the view, or `None`.
        """
        if not hasattr(self, "_paginator"):
            if self.pagination_class is None:
                self._paginator = None
            else:
                self._paginator = self.pagination_class()
        return self._paginator

    def paginate_queryset(self, queryset, **kwargs):
        """
        Return a single page of results, or `None` if pagination is disabled.
        """
        if self.paginator is None:
            return None
        return self.paginator.paginate_queryset(queryset, self.request, view=self, **kwargs)


class InputMixin:
    """
    Take an input operation argument and return the data.

    This mixin is pretty limited, nested enums under other input types will not be
    converted into their value. FIX/CHANGE
    """

    input_arg = "input"
    convert_enums = True

    def get_input_arg(self):
        return self.input_arg

    def get_input_data(self):
        input_data = self.operation_kwargs.get(self.input_arg, {}).copy()
        # Get and convert any translated enums into their values
        # May be able to be moved to a serializer save point for models?
        # As having access to the enum may be useful in other contexts
        if self.convert_enums:
            if isinstance(input_data, dict):
                self.process_dict(input_data)

            if isinstance(input_data, list):
                for data in input_data:
                    if isinstance(data, dict):
                        data = self.process_dict(data)

        return input_data

    def process_dict(self, input_data):
        for key, value in input_data.items():
            if isinstance(value, enum.Enum):
                # Get and set actual value of enum
                input_data[key] = value.value


class CreateModelMixin(InputMixin):
    def create(self, parent, *args, **kwargs):
        serializer = self.get_serializer(data=self.get_input_data())
        valid = serializer.is_valid(raise_exception=False)
        if valid:
            obj = self.perform_create(serializer)
        else:
            # This will break shit
            obj = None
        return dict(success=valid, object=obj, errors=serializer.errors)

    def perform_create(self, serializer):
        return serializer.save()


class DetailModelMixin:
    lookup_arg = None
    lookup_field = "id"

    def get_lookup_arg(self):
        return self.config.get("lookup_arg", self.lookup_arg)

    def get_lookup_field(self):
        return self.config.get("lookup_field", self.lookup_field)

    def get_lookup_operation_data(self):
        if self.config.get("reference", False):
            return self.reference_kwargs
        return self.operation_kwargs

    def get_lookup_filter_kwargs(self):
        # Perform the lookup filtering.
        lookup_arg = self.get_lookup_arg() or self.get_lookup_field()

        operation_data = self.get_lookup_operation_data()

        assert lookup_arg in operation_data, (
            "Expected resolver %s to be called with an argument "
            'named "%s". Fix your query arguments, or set the `.lookup_field` '
            "attribute on the resolver correctly." % (self.__class__.__name__, lookup_arg)
        )
        return {self.get_lookup_field(): operation_data[lookup_arg]}

    def get_object(self):
        """
        Returns a singular object as configured by the resolver

        You may want to override this if you need to provide non-standard
        queryset lookups. Eg if objects are referenced using multiple
        arguments.
        """
        queryset = self.get_queryset()

        filter_kwargs = self.get_lookup_filter_kwargs()

        try:
            obj = queryset.get(**filter_kwargs)
        except queryset.model.DoesNotExist:
            raise exceptions.NotFoundException()

        # TODO: handle no object found for field, raise exception that always is caught and returns null for field?

        # May raise a permission denied
        if obj:
            self.check_object_permissions(self.request, obj)

        return obj

    # TODO: move out into its own mixin?
    def retrieve(self, parent, *args, **kwargs):
        return self.get_object()


class UpdateModelMixin(DetailModelMixin):
    """
    Update a model instance.
    """

    def update(self, request, *args, **kwargs):
        partial = kwargs.pop("partial", False)
        instance = self.get_object()
        serializer = self.get_serializer(instance, data=self.get_input_data(), partial=partial)
        valid = serializer.is_valid(raise_exception=False)

        if valid:
            instance = self.perform_update(serializer)

        if getattr(instance, "_prefetched_objects_cache", None):
            # If 'prefetch_related' has been applied to a queryset, we need to
            # forcibly invalidate the prefetch cache on the instance.
            instance._prefetched_objects_cache = {}

        return dict(success=valid, object=instance, errors=serializer.errors)

    def perform_update(self, serializer):
        return serializer.save()

    def partial_update(self, request, *args, **kwargs):
        """
        Resolver method to use when you want to configure the serializer for partial updates
        """
        kwargs["partial"] = True
        return self.update(request, *args, **kwargs)


class DestroyModelMixin(DetailModelMixin):
    """
    Destroy a model instance.
    """

    def destroy(self, request, *args, **kwargs):
        instance = self.get_object()
        instance_id = instance.id
        count = False
        errors = {}
        try:
            count, _ = self.perform_destroy(instance)
        except (IntegrityError, ProtectedError) as e:
            errors = {"destroy": [str(e)]}
        return {"success": bool(count), "errors": errors, "object": {"id": instance_id}}

    def perform_destroy(self, instance):
        return instance.delete()



from ariadne_extended.resolvers.model import ModelResolver

from .mixins import RelayModelMixin
from ariadne import FallbackResolversSetter
from .types import page_info
from graphql.type import GraphQLField

__resolvers__ = ["page_info_resolver"]


@page_info.field("count")
def resolve_page_info_count(parent, info, *args, **kwargs):
    page = parent.get("cursor", None)
    return page.paginator.queryset.count()


def resolve_page_info(parent, info, *args, **kwargs) -> dict:
    page = parent.get("pageInfo", None)
    try:
        end_cursor = page.cursor(page[-1])
    except IndexError:
        end_cursor = None
    try:
        start_cursor = page.cursor(page[0])
    except IndexError:
        start_cursor = None
    return {
        "hasNextPage": page.has_next,
        "hasPreviousPage": page.has_previous,
        "startCursor": start_cursor,
        "endCursor": end_cursor,
        "has_next_page": page.has_next,
        "has_previous_page": page.has_previous,
        "start_cursor": start_cursor,
        "end_cursor": end_cursor,
        "cursor": page,
    }


class CustomPageInfoResolver(FallbackResolversSetter):
    def add_resolver_to_field(self, name: str, field_object: GraphQLField) -> None:
        if name == "pageInfo" and field_object.resolve is None:
            field_object.resolve = resolve_page_info


page_info_resolver = CustomPageInfoResolver()


class RelayModelResolver(RelayModelMixin, ModelResolver):
    """
    Model resolver that provides cursor based pagination
    """

    pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ariadne_extended',
 'ariadne_extended.contrib.waffle_graph',
 'ariadne_extended.cursor_pagination',
 'ariadne_extended.filters',
 'ariadne_extended.graph_loader',
 'ariadne_extended.payload',
 'ariadne_extended.resolvers',
 'ariadne_extended.utils',
 'ariadne_extended.uuid']

package_data = \
{'': ['*']}

install_requires = \
['ariadne>=0.12.0,<0.13.0',
 'django>=2.2,<=3.2',
 'djangorestframework>=3.7,<4.0',
 'pyhumps>=1.6.1,<2.0.0']

setup_kwargs = {
    'name': 'ariadne-extended',
    'version': '0.2.0',
    'description': 'Ariadne contrib library for working with Django and GraphQL',
    'long_description': "# Ariadne Extended\n\nAriadne comes with a few contrib modules to support integration with Django. This module acts as an additional Django contrib module to reduce some of the boilerplate when integrating with Django.\n\nIt copies and follows some of the conventions defined in Django Rest Framework, effort will be made to provide API compatibility with certain DRF extension modules that are relevant.\n\nThe overall goal is to provide often re-used modules and GrpahQL schema, so one doesn't have to re-create them per project.\n\n## Supported DRF modules\n\n* Pagination classes\n* Permissions classes\n* Serializer classes ( may not removed and made generic )\n* Throttling\n* django-filters filter backend\n\n## Supported Django versions\n\n**2.2.*, 3.0.\\***\n\n\n### `ariadne_extended.graph_loader`\n\nauto schema, resolver and types loader \n\nSearches for `types` and `resolvers` modules as well as any `.graphql` files defined within any installed Django application.\n\nOnce found they can be used to build the final schema and resolver solution for your ariadne application.\n\n### `ariadne_extended.payload`\nSchema, types and resolvers for the graphql `Payload` interface.\n\nCurrently the `FieldErrors` are highly coupled to Django rest framework field validation exceptions.\n\n### `ariadne_extended.cursor_pagination`\nRename to relay_pagination maybe?\n\nContains the `PageInfo` graphql type and `Connection` interface for utilizing cursor based pagination.\n\nTODO: Should we provide a input type for cursor pagination instead of just copying pagination args info paginate-able list fields??\n\n### `ariadne_extended.filters`\nFilter backend interface to pass filter arguments to django-filter.\n\n### `ariadne_extended.resolvers`\nABC for Class Based Resolvers and model resolvers that utilize DRF serializers for saving data. This is likely to change in the future.\n\n### `ariadne_extended.uuid`\nDRF UUID field scalar for use with models that may use a UUID as their primary key, or other UUID fields.\n\n## Contrib\n\n### `django-waffle` in `ariadne_extended.contrib.waffle_graph`\n\nAdds resolvers, schema and types that can be utilized to query the any waffle flags, samples and switches.\n\n\n# Checklist\n- [ ] Get rid of dependency on DRF?\n- [ ] Investigate the need for a serializer ( nested data reliance? )\n- [x] Organize code into multiple Django apps to select desired componentry\n- [ ] Documentation and examples\n- [ ] Better support of lists of enums when used with django-filters ( currently expects comma sep list string, not a list of enums from input field resolver )\n- [x] License and make public\n- [ ] Deployment automation\n- [ ] Mixins are highly coupled to serializers, should they still be?\n",
    'author': 'Patrick Forringer',
    'author_email': 'pforringer@patriotrc.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/patriotresearch/ariadne-extended',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
